/*
 * To change this license header, choose License Headers in Project Properties.
 * To change this template file, choose Tools | Templates
 * and open the template in the editor.
 */
package mentionmatch;

import java.io.PrintStream;
import java.util.Arrays;

/**
 *
 * @author gbl
 */

class MatrixComparison {

    private class Coord implements Comparable { 

        int x; int y;
        private Coord(int bestX, int bestY) {
            x=bestX;
            y=bestY;
        }

        @Override
        public int compareTo(Object o2) {
            if (o2 instanceof Coord) {
                return (this).x - ((Coord)o2).x;
            }
            throw new IllegalArgumentException("Coords can only be compared to coords");
        }
    }

    ListOfMentions measure, reference;
    int meaStart, meaEnd, refStart, refEnd;
    float xAxisVal[], yAxisVal[];
    float[][] matrix;
    Coord[] matches;

    MatrixComparison(ListOfMentions measure, int measureStart, int measureEnd, ListOfMentions reference, int referenceStart, int referenceEnd) {
        this.measure=measure;
        this.meaStart=measureStart;
        this.meaEnd=measureEnd;
        
        this.reference=reference;
        this.refStart=referenceStart;
        this.refEnd=referenceEnd;
        
        matrix=new float[meaEnd-meaStart+1][];
        for (int i=0; i<meaEnd-meaStart+1; i++) {
            matrix[i]=new float[refEnd-refStart+1];
        }

        xAxisVal=new float[meaEnd-meaStart+1];
        yAxisVal=new float[refEnd-refStart+1];

        matches=null;
    }

    void process() {

        System.out.printf("comparing measure indexes %d-%d to reference %d-%d\n", meaStart, meaEnd, refStart, refEnd);

        // Create a matrix that has, as its X axis, mea positions normalized to [0,1],
        // and as its Y axis, ref positions, normalized as well. For example, the
        // measurement list [0, 4, 5, 10] and the reference list [0, 6, 9, 12]
        // will lead to
        //              0.0     0.4     0.5     1.0
        //  0.0
        //  0.5
        //  0.75
        //  1.0
        //
        // then fill it with the distances:
        //              0.0     0.4     0.5     1.0
        //  0.0         0.0     0.4     0.5     1.0
        //  0.5         0.5     0.1     0.0     0.5
        //  0.75        0.75    0.35    0.25    0.25
        //  1.0         1.0     0.6     0.5     0.0
        // Actually we don't store the axes, just the contents.
       
        for (int x=0; x<meaEnd-meaStart+1; x++)
            xAxisVal[x] = normalize(measure.getMentionAt(x+meaStart).peakPos, measure.getMentionAt(meaStart).peakPos, measure.getMentionAt(meaEnd).peakPos);
        for (int y=0; y<refEnd-refStart+1; y++)
            yAxisVal[y] = normalize(reference.getMentionAt(y+refStart).peakPos, reference.getMentionAt(refStart).peakPos, reference.getMentionAt(refEnd).peakPos);
            
        for (int x=0; x<meaEnd-meaStart+1; x++) {
            for (int y=0; y<refEnd-refStart+1; y++) {
                matrix[x][y]=Math.abs(yAxisVal[y]-xAxisVal[x]);
            }
        }
        dumpMatrix("filled distances");
        
        // Now mark possible matches; there can be only one match per row and column
        // (so the number of matches is the minimum of rows/columns), each placement
        // goes to the cell that has the lowest value in the matrix and isn't already
        // blocked my another match in the same row or column. This leads to
        //              0.0     0.4     0.5     1.0
        //  0.0        *0.0     0.4     0.5     1.0
        //  0.5         0.5     0.1    *0.0     0.5
        //  0.75        0.75   *0.35    0.25    0.25
        //  1.0         1.0     0.6     0.5    *0.0
        
        
        int numberOfMatches = Math.min(refEnd-refStart+1, meaEnd-meaStart+1);
        matches=new Coord[numberOfMatches];
        boolean blockedX[]=new boolean[meaEnd-meaStart+1];
        boolean blockedY[]=new boolean[refEnd-refStart+1];

        for (int i=0; i<numberOfMatches; i++) {
            float lowestValue = Float.MAX_VALUE;        // ensure we find a lower matrix value
            int bestX=-1, bestY=-1;
            for (int x=0; x<meaEnd-meaStart+1; x++) {
                if (blockedX[x])
                    continue;
                for (int y=0; y<refEnd-refStart+1; y++) {
                    if (blockedY[y])
                        continue;
                    if (matrix[x][y] < lowestValue) {
                        bestX=x;
                        bestY=y;
                        lowestValue = matrix[x][y];
                    }
                }
            }
            if (bestX==-1 || bestY==-1) {
                System.err.println("Finding a matrix value didn't actually find anything\n");
                System.exit(1);
            }
            matches[i]=new Coord(bestX, bestY);
            blockedX[bestX]=true;
            blockedY[bestY]=true;
        }
        dumpMatrix("tried to match");

/* Forget about this now; this doesn't work.

        // There may still be an occasional swap, like with 0.1/0.0/0.35/0.25 above.
        // On the measurements, this looks like this:
        // M----------------U---V-----------------N
        // M--------------------W---X-------------N
        // This may either be due to inconsistent stretching (in which case UX/VW
        // should be turned into UW/VX), or one new wire break (U) and an old one
        // that wasn't recognized (X). We assume it's inconsistent stretching and
        // swap the matches if the rations AU/AW, UV/WX and VB/XB are close to 1.0,
        // and assume it's new/old breaks if they aren't.
        // Note we can also have something like
        // M---------------A-B-C-D----------------N
        // M-----------P-Q-R-S--------------------N
        // where we get AR/BS/CQ/DP but want AP/BQ/CR/DS.
        
        Arrays.sort(matches);
        float maxAllowedStretch = ratio(matches[matches.length-1].x, matches[matches.length-1].y)*1.1f;      // allow 10% stretch
        for (int i=0; i<matches.length-1; i++) {
            if (matches[i].y > matches[i+1].y) {    // i=B, i+1=C in above case
                // find the last pair that has decrementing Y (DP)
                int decrementingLength=1;
                int minInvolvedY=matches[i+1].y;
                while (i+decrementingLength < matches.length-1 
                    && matches[i].y > matches[i+decrementingLength+1].y) {
                    if (minInvolvedY > matches[i+decrementingLength+1].y) {
                        minInvolvedY = matches[i+decrementingLength+1].y;
                    }
                    decrementingLength++;
                }
                // here, i should index B, and decrementingLength should be 2 so i+DL indexes D
                // next, count how many elements we need to go back from S to find P. As matches
                // is sorted by X, and we don't have a sorted array of Y's, count how many
                // Ys that are actually referenced are in [P, S] - here, 4.
                int matchingYs=0;
                for (int j=0; j<matches.length; j++) {
                    if (matches[j].y >= minInvolvedY
                    &&  matches[j].y <= matches[i].y) {
                         matchingYs++;
                    }
                }
                int lastChainX=i+decrementingLength;                // find D
                int firstChainX=i+decrementingLength+1-matchingYs;  // find A
                if (ratio(xAxisVal[matches[firstChainX].x]-0.0f, yAxisVal[matches[lastChainX].y]-0.0f) < maxAllowedStretch      // compare MA to MP
                &&  ratio(xAxisVal[matches[lastChainX].x]-xAxisVal[matches[firstChainX].x], yAxisVal[matches[i].y]-yAxisVal[matches[lastChainX].y]) < maxAllowedStretch // compare AD to PS
                &&  ratio(1.0f-xAxisVal[matches[lastChainX].x], 1.0f-yAxisVal[matches[i].y]) < maxAllowedStretch) {
                    System.out.printf("I should switch between %d and %d but don't know how yet\n", firstChainX, lastChainX);
                    int[] resorted = new int[lastChainX-firstChainX+1];
                    // switch: AR BS CQ DP to AP BQ CR DS
                } else {
                    // remove match CQ, DP
                    for (int j=1; j<=decrementingLength; j++) {
                        matches[i+j].x=matches[i+j].y=-1;
                    }
                }
                i+=decrementingLength;      // so it starts after current stretch on next iteration
            }
        }
*/

        Arrays.sort(matches);
        for (int i=0; i<matches.length-1; i++) {
            if (matches[i+1].x==-1 || matches[i+1].y==-1)
                break;
            if (matches[i].y > matches[i+1].y) {
                // we have a backwards match; eliminate the worse one of the both
                // by copying the following elements back and replacing the
                // last one with invalid (-1,-1)
                float myDistance=matrix[matches[i].x][matches[i].y];
                float nextDistance=matrix[matches[i+1].x][matches[i+1].y];
                
                for (int j=(myDistance > nextDistance ? i : i+1); j<matches.length-1; j++) {
                    matches[j]=matches[j+1];
                }
                matches[matches.length-1]=new Coord(-1, -1);
                i=-1;    // and redo from start, the previous match might be broken as well
            }
        }
    }
        

    /**
     * Normalize a value X in an interval from A to B to a value from 0.0 to 1.0
     * depending on the position between A and B.
     * @param x the value to normalize
     * @param a the interval start
     * @param b the interval end
     * @return the normalized value between 0 and 1
     */
    private float normalize(int x, int a, int b) {
        return (float)(x-a)/(float)(b-a);
    }
    
    private float ratio(float a, float b) {
        float x = ( a > b ? a / b : b / a );
        System.out.printf("rrrrrr ratio of %6.2f and %6.2f is %6.2f\n", a, b, x);
        return x;
    }
    
    private void dumpMatrix(String header) {
        System.out.println("============ "+header+" ================");
        System.out.printf("%8s", "");
        for (int x=0; x<meaEnd-meaStart+1; x++) {
            System.out.printf("%10.5f ", xAxisVal[x]);
        }
        System.out.println();
        
        for (int y=0; y<refEnd-refStart+1; y++) {
            System.out.printf("%8.5f", yAxisVal[y]);
            for (int x=0; x<meaEnd-meaStart+1; x++) {
                System.out.printf("%10.5f", matrix[x][y]);
                boolean isMatch=false;
                if (matches != null) {
                    for (int i=0; i<matches.length; i++) {
                        if (x==matches[i].x && y==matches[i].y) {
                            isMatch=true;
                            break;
                        }
                    }
                }
                System.out.print(isMatch ? '*' : ' ');
            }
            System.out.println();
        }
        System.out.println("---------------------------------------");
    }

    void print(PrintStream out) {
        dumpMatrix("final comparison matrix");
        
        System.out.println("+++ matching mentions");
        for (int i=0; i<matches.length; i++) {
            if (matches[i].x != -1 && matches[i].y != -1) {
                System.out.printf("Measure index %3d type %-20s at %10d matches reference %3d type %-20s at %10d with delta %5.3f\n", 
                        matches[i].x+meaStart, measure.getMentionAt(matches[i].x+meaStart).description, measure.getMentionAt(matches[i].x+meaStart).peakPos,
                        matches[i].y+refStart, reference.getMentionAt(matches[i].y+refStart).description, reference.getMentionAt(matches[i].y+refStart).peakPos,
                        matrix[matches[i].x][matches[i].y]
                        );
            }
        }
        
        System.out.println("+++ new mentions");
        for (int x=0; x<=meaEnd-meaStart; x++) {
            boolean isMatch=false;
            for (int i=0; i<matches.length; i++) {
                if (matches[i].x == x) {
                    isMatch=true;
                    break;
                }
            }
            if (!isMatch) {
                System.out.printf("Measure index %3d type %-20s at %10d is new\n", x+meaStart, measure.getMentionAt(x+meaStart).description, measure.getMentionAt(x+meaStart).peakPos);
            }
        }

        System.out.println("+++ in reference but not found");
        for (int y=0; y<=refEnd-refStart; y++) {
            boolean isMatch=false;
            for (int i=0; i<matches.length; i++) {
                if (matches[i].y == y) {
                    isMatch=true;
                    break;
                }
            }
            if (!isMatch) {
                System.out.printf("Reference index %3d type %-20s at %10d is new\n", y+refStart, reference.getMentionAt(y+refStart).description, reference.getMentionAt(y+refStart).peakPos);
            }
        }
    }
}
